"""
    get_permutation_from_translation(translation::Tuple{Int,Int}, stdBasis::StandardBasis, d)

Return `Permutations.Permutation` of translation for ``d^2`` dimensional vector of coordinates in Bell basis `standardBasis`.
"""
function get_permutation_from_translation(translation::Tuple{Int,Int}, stdBasis::StandardBasis, d)

    myDicts = create_dictionary_from_basis(stdBasis)
    dict = myDicts[1] #keys are tuples 
    reverseDict = myDicts[2] #keys are indices

    D = length(myDicts[1])
    t = collect(translation)

    perm = zeros(D)

    for i in 1:D

        oldCoords = collect(reverseDict[i])
        newCoords = mod.(oldCoords + t, d)
        permIndex = dict[Tuple(newCoords)]

        perm[i] = permIndex

    end

    return Permutation(perm)

end

"""
    get_permutation_from_momentuminversion(stdBasis::StandardBasis, d)

Return `Permutations.Permutation` of momentum inversion for ``d^2`` dimensional vector of coordinates in Bell basis `standardBasis`.
"""
function get_permutation_from_momentuminversion(stdBasis::StandardBasis, d)

    #load dictionary
    myDicts = create_dictionary_from_basis(stdBasis)
    dict = myDicts[1] #keys are tuples 
    reverseDict = myDicts[2] #keys are indices

    D = length(myDicts[1])

    perm = zeros(D)

    for i in 1:D

        oldCoords = collect(reverseDict[i])
        newCoords = mod.([-oldCoords[1], oldCoords[2]], d)
        permIndex = dict[Tuple(newCoords)]

        perm[i] = permIndex

    end

    return Permutation(perm)

end

"""
    get_permutation_from_quarterrotation(stdBasis::StandardBasis, d)

Return `Permutations.Permutation` of qurater rotation for ``d^2`` dimensional vector of coordinates in Bell basis `standardBasis`.
"""
function get_permutation_from_quarterrotation(stdBasis::StandardBasis, d)

    #load dictionary
    myDicts = create_dictionary_from_basis(stdBasis)
    dict = myDicts[1] #keys are tuples 
    reverseDict = myDicts[2] #keys are indices

    D = length(myDicts[1])

    perm = zeros(D)

    for i in 1:D

        oldCoords = collect(reverseDict[i])
        newCoords = mod.([oldCoords[2], -oldCoords[1]], d)
        permIndex = dict[Tuple(newCoords)]

        perm[i] = permIndex

    end

    return Permutation(perm)

end

"""
    get_permutation_from_verticalshear(stdBasis::StandardBasis, d)

Return `Permutations.Permutation`` of vertical shear for ``d^2`` dimensional vector of coordinates in Bell basis`standardBasis`.
"""
function get_permutation_from_verticalshear(stdBasis::StandardBasis, d)

    #load dictionary
    myDicts = create_dictionary_from_basis(stdBasis)
    dict = myDicts[1] #keys are tuples 
    reverseDict = myDicts[2] #keys are indices

    D = length(myDicts[1])

    perm = zeros(D)

    for i in 1:D

        oldCoords = collect(reverseDict[i])
        newCoords = mod.([oldCoords[1] + oldCoords[2], oldCoords[2]], d)
        permIndex = dict[Tuple(newCoords)]

        perm[i] = permIndex

    end

    return Permutation(perm)

end

"""
    generate_symmetries(stdBasis::StandardBasis, d, orderLimit=0)

Return array of `Permutations.Permutation` of all symmetries up to order `orderLimit` in `d` dimensions generated by the generators represented in `standardBasis`.
"""
function generate_symmetries(stdBasis::StandardBasis, d, orderLimit=0)::Array{Permutation}

    allSymmetries = Permutation[]
    #Get all translations
    for i in 0:(d-1)
        for j in 0:(d-1)
            push!(allSymmetries, get_permutation_from_translation((i, j), stdBasis, d))
        end
    end

    generators = (
        get_permutation_from_quarterrotation(stdBasis, d),
        get_permutation_from_momentuminversion(stdBasis, d),
        get_permutation_from_verticalshear(stdBasis, d)
    )

    findNew = true
    order = 1

    while findNew

        foundSims = Permutation[]

        for x in generators
            generatedSyms = map(s -> s * x, allSymmetries)
            newSyms = filter(e -> !(e in allSymmetries), generatedSyms)
            append!(foundSims, newSyms)
        end

        if length(foundSims) > 0
            append!(allSymmetries, foundSims)
        end

        order += 1

        findNew = (length(foundSims) > 0 && (orderLimit >= order || orderLimit == 0))

        unique!(allSymmetries)

    end

    return allSymmetries

end

"""
    get_symcoords(coords::Array{Float64,1}, symPermutations::Array{Permutation})

Return array containing all symmetric Bell coordinates of given symmetries `symPermutations` applied to Bell coordinates `coords`.
"""
function get_symcoords(coords::Array{Float64,1}, symPermutations::Array{Permutation})

    symCoords = map(x -> Matrix(x) * coords, symPermutations)
    return symCoords

end